@@ -481,11 +481,12 @@ static int do_balance(const char *path, struct btrfs_ioctl_balance_args *args,
 			"There may be more info in syslog - try dmesg | tail\n");
 			ret = 1;
 		}
+	} else if (ret > 0) {
+		error("balance: %s", btrfs_err_str(ret));
 	} else {
 		printf("Done, had to relocate %llu out of %llu chunks\n",
 		       (unsigned long long)args->stat.completed,
 		       (unsigned long long)args->stat.considered);
-		ret = 0;
 	}
 
 out:
@@ -508,7 +509,7 @@ static const char * const cmd_balance_start_usage[] = {
 	"-m[filters]    act on metadata chunks",
 	"-s[filters]    act on system chunks (only under -f)",
 	"-v             be verbose",
-	"-f             force reducing of metadata integrity",
+	"-f             force a reduction of metadata integrity",
 	"--full-balance do not print warning and do not delay start",
 	"--background|--bg",
 	"               run the balance as a background process",
@@ -585,12 +586,12 @@ static int cmd_balance_start(int argc, char **argv)
 			background = 1;
 			break;
 		default:
-			usage(cmd_balance_start_usage);
+			usage_unknown_option(cmd_balance_start_usage, argv);
 		}
 	}
 
 	if (check_argc_exact(argc - optind, 1))
-		usage(cmd_balance_start_usage);
+		return 1;
 
 	/*
 	 * allow -s only under --force, otherwise do with system chunks
@@ -690,7 +691,7 @@ static int cmd_balance_pause(int argc, char **argv)
 	clean_args_no_options(argc, argv, cmd_balance_pause_usage);
 
 	if (check_argc_exact(argc - optind, 1))
-		usage(cmd_balance_pause_usage);
+		return 1;
 
 	path = argv[optind];
 
@@ -728,7 +729,7 @@ static int cmd_balance_cancel(int argc, char **argv)
 	clean_args_no_options(argc, argv, cmd_balance_cancel_usage);
 
 	if (check_argc_exact(argc - optind, 1))
-		usage(cmd_balance_cancel_usage);
+		return 1;
 
 	path = argv[optind];
 
@@ -767,7 +768,7 @@ static int cmd_balance_resume(int argc, char **argv)
 	clean_args_no_options(argc, argv, cmd_balance_resume_usage);
 
 	if (check_argc_exact(argc - optind, 1))
-		usage(cmd_balance_resume_usage);
+		return 1;
 
 	path = argv[optind];
 
@@ -849,12 +850,12 @@ static int cmd_balance_status(int argc, char **argv)
 			verbose = 1;
 			break;
 		default:
-			usage(cmd_balance_status_usage);
+			usage_unknown_option(cmd_balance_status_usage, argv);
 		}
 	}
 
 	if (check_argc_exact(argc - optind, 1))
-		usage(cmd_balance_status_usage);
+		return 1;
 
 	path = argv[optind];
 
