@@ -26,15 +26,13 @@
 #include "commands.h"
 #include "utils.h"
 #include "help.h"
+#include "rescue.h"
 
 static const char * const rescue_cmd_group_usage[] = {
 	"btrfs rescue <command> [options] <path>",
 	NULL
 };
 
-int btrfs_recover_chunk_tree(const char *path, int verbose, int yes);
-int btrfs_recover_superblocks(const char *path, int verbose, int yes);
-
 static const char * const cmd_rescue_chunk_recover_usage[] = {
 	"btrfs rescue chunk-recover [options] <device>",
 	"Recover the chunk tree by scanning the devices one by one.",
@@ -64,14 +62,14 @@ static int cmd_rescue_chunk_recover(int argc, char *argv[])
 		case 'v':
 			verbose = 1;
 			break;
-		case 'h':
 		default:
-			usage(cmd_rescue_chunk_recover_usage);
+			usage_unknown_option(cmd_rescue_chunk_recover_usage,
+					argv);
 		}
 	}
 
 	if (check_argc_exact(argc - optind, 1))
-		usage(cmd_rescue_chunk_recover_usage);
+		return 1;
 
 	file = argv[optind];
 
@@ -134,11 +132,12 @@ static int cmd_rescue_super_recover(int argc, char **argv)
 			yes = 1;
 			break;
 		default:
-			usage(cmd_rescue_super_recover_usage);
+			usage_unknown_option(cmd_rescue_super_recover_usage,
+					argv);
 		}
 	}
 	if (check_argc_exact(argc - optind, 1))
-		usage(cmd_rescue_super_recover_usage);
+		return 1;
 
 	dname = argv[optind];
 	ret = check_mounted(dname);
@@ -172,7 +171,7 @@ static int cmd_rescue_zero_log(int argc, char **argv)
 	clean_args_no_options(argc, argv, cmd_rescue_zero_log_usage);
 
 	if (check_argc_exact(argc, 2))
-		usage(cmd_rescue_zero_log_usage);
+		return 1;
 
 	devname = argv[optind];
 	ret = check_mounted(devname);
@@ -224,7 +223,7 @@ static int cmd_rescue_fix_device_size(int argc, char **argv)
 	clean_args_no_options(argc, argv, cmd_rescue_fix_device_size_usage);
 
 	if (check_argc_exact(argc, 2))
-		usage(cmd_rescue_fix_device_size_usage);
+		return 1;
 
 	devname = argv[optind];
 	ret = check_mounted(devname);
