@@ -36,54 +36,11 @@
 #include "volumes.h"
 #include "extent_io.h"
 #include "help.h"
-
-#define HEADER_MAGIC		0xbd5c25e27295668bULL
-#define MAX_PENDING_SIZE	(256 * 1024)
-#define BLOCK_SIZE		1024
-#define BLOCK_MASK		(BLOCK_SIZE - 1)
-
-#define COMPRESS_NONE		0
-#define COMPRESS_ZLIB		1
+#include "image/metadump.h"
+#include "image/sanitize.h"
 
 #define MAX_WORKER_THREADS	(32)
 
-struct meta_cluster_item {
-	__le64 bytenr;
-	__le32 size;
-} __attribute__ ((__packed__));
-
-struct meta_cluster_header {
-	__le64 magic;
-	__le64 bytenr;
-	__le32 nritems;
-	u8 compress;
-} __attribute__ ((__packed__));
-
-/* cluster header + index items + buffers */
-struct meta_cluster {
-	struct meta_cluster_header header;
-	struct meta_cluster_item items[];
-} __attribute__ ((__packed__));
-
-#define ITEMS_PER_CLUSTER ((BLOCK_SIZE - sizeof(struct meta_cluster)) / \
-			   sizeof(struct meta_cluster_item))
-
-struct fs_chunk {
-	u64 logical;
-	u64 physical;
-	/*
-	 * physical_dup only store additonal physical for BTRFS_BLOCK_GROUP_DUP
-	 * currently restore only support single and DUP
-	 * TODO: modify this structure and the function related to this
-	 * structure for support RAID*
-	 */
-	u64 physical_dup;
-	u64 bytes;
-	struct rb_node l;
-	struct rb_node p;
-	struct list_head list;
-};
-
 struct async_work {
 	struct list_head list;
 	struct list_head ordered;
@@ -120,18 +77,11 @@ struct metadump_struct {
 	int compress_level;
 	int done;
 	int data;
-	int sanitize_names;
+	enum sanitize_mode sanitize_names;
 
 	int error;
 };
 
-struct name {
-	struct rb_node n;
-	char *val;
-	char *sub;
-	u32 len;
-};
-
 struct mdrestore_struct {
 	FILE *in;
 	FILE *out;
@@ -169,11 +119,11 @@ static struct extent_buffer *alloc_dummy_eb(u64 bytenr, u32 size);
 
 static void csum_block(u8 *buf, size_t len)
 {
-	u8 result[BTRFS_CRC32_SIZE];
+	u8 result[btrfs_csum_sizes[BTRFS_CSUM_TYPE_CRC32]];
 	u32 crc = ~(u32)0;
 	crc = crc32c(crc, buf + BTRFS_CSUM_SIZE, len - BTRFS_CSUM_SIZE);
 	btrfs_csum_final(crc, result);
-	memcpy(buf, result, BTRFS_CRC32_SIZE);
+	memcpy(buf, result, btrfs_csum_sizes[BTRFS_CSUM_TYPE_CRC32]);
 }
 
 static int has_name(struct btrfs_key *key)
@@ -192,35 +142,6 @@ static int has_name(struct btrfs_key *key)
 	return 0;
 }
 
-static char *generate_garbage(u32 name_len)
-{
-	char *buf = malloc(name_len);
-	int i;
-
-	if (!buf)
-		return NULL;
-
-	for (i = 0; i < name_len; i++) {
-		char c = rand_range(94) + 33;
-
-		if (c == '/')
-			c++;
-		buf[i] = c;
-	}
-
-	return buf;
-}
-
-static int name_cmp(struct rb_node *a, struct rb_node *b, int fuzz)
-{
-	struct name *entry = rb_entry(a, struct name, n);
-	struct name *ins = rb_entry(b, struct name, n);
-	u32 len;
-
-	len = min(ins->len, entry->len);
-	return memcmp(ins->val, entry->val, len);
-}
-
 static int chunk_cmp(struct rb_node *a, struct rb_node *b, int fuzz)
 {
 	struct fs_chunk *entry = rb_entry(a, struct fs_chunk, l);
@@ -338,243 +259,6 @@ static u64 logical_to_physical(struct mdrestore_struct *mdres, u64 logical,
 	return fs_chunk->physical + offset;
 }
 
-
-static char *find_collision(struct metadump_struct *md, char *name,
-			    u32 name_len)
-{
-	struct name *val;
-	struct rb_node *entry;
-	struct name tmp;
-	unsigned long checksum;
-	int found = 0;
-	int i;
-
-	tmp.val = name;
-	tmp.len = name_len;
-	entry = tree_search(&md->name_tree, &tmp.n, name_cmp, 0);
-	if (entry) {
-		val = rb_entry(entry, struct name, n);
-		free(name);
-		return val->sub;
-	}
-
-	val = malloc(sizeof(struct name));
-	if (!val) {
-		error("cannot sanitize name, not enough memory");
-		free(name);
-		return NULL;
-	}
-
-	memset(val, 0, sizeof(*val));
-
-	val->val = name;
-	val->len = name_len;
-	val->sub = malloc(name_len);
-	if (!val->sub) {
-		error("cannot sanitize name, not enough memory");
-		free(val);
-		free(name);
-		return NULL;
-	}
-
-	checksum = crc32c(~1, val->val, name_len);
-	memset(val->sub, ' ', name_len);
-	i = 0;
-	while (1) {
-		if (crc32c(~1, val->sub, name_len) == checksum &&
-		    memcmp(val->sub, val->val, val->len)) {
-			found = 1;
-			break;
-		}
-
-		if (val->sub[i] == 127) {
-			do {
-				i++;
-				if (i >= name_len)
-					break;
-			} while (val->sub[i] == 127);
-
-			if (i >= name_len)
-				break;
-			val->sub[i]++;
-			if (val->sub[i] == '/')
-				val->sub[i]++;
-			memset(val->sub, ' ', i);
-			i = 0;
-			continue;
-		} else {
-			val->sub[i]++;
-			if (val->sub[i] == '/')
-				val->sub[i]++;
-		}
-	}
-
-	if (!found) {
-		warning(
-"cannot find a hash collision for '%.*s', generating garbage, it won't match indexes",
-			val->len, val->val);
-		for (i = 0; i < name_len; i++) {
-			char c = rand_range(94) + 33;
-
-			if (c == '/')
-				c++;
-			val->sub[i] = c;
-		}
-	}
-
-	tree_insert(&md->name_tree, &val->n, name_cmp);
-	return val->sub;
-}
-
-static void sanitize_dir_item(struct metadump_struct *md, struct extent_buffer *eb,
-			      int slot)
-{
-	struct btrfs_dir_item *dir_item;
-	char *buf;
-	char *garbage;
-	unsigned long name_ptr;
-	u32 total_len;
-	u32 cur = 0;
-	u32 this_len;
-	u32 name_len;
-	int free_garbage = (md->sanitize_names == 1);
-
-	dir_item = btrfs_item_ptr(eb, slot, struct btrfs_dir_item);
-	total_len = btrfs_item_size_nr(eb, slot);
-	while (cur < total_len) {
-		this_len = sizeof(*dir_item) +
-			btrfs_dir_name_len(eb, dir_item) +
-			btrfs_dir_data_len(eb, dir_item);
-		name_ptr = (unsigned long)(dir_item + 1);
-		name_len = btrfs_dir_name_len(eb, dir_item);
-
-		if (md->sanitize_names > 1) {
-			buf = malloc(name_len);
-			if (!buf) {
-				error("cannot sanitize name, not enough memory");
-				return;
-			}
-			read_extent_buffer(eb, buf, name_ptr, name_len);
-			garbage = find_collision(md, buf, name_len);
-		} else {
-			garbage = generate_garbage(name_len);
-		}
-		if (!garbage) {
-			error("cannot sanitize name, not enough memory");
-			return;
-		}
-		write_extent_buffer(eb, garbage, name_ptr, name_len);
-		cur += this_len;
-		dir_item = (struct btrfs_dir_item *)((char *)dir_item +
-						     this_len);
-		if (free_garbage)
-			free(garbage);
-	}
-}
-
-static void sanitize_inode_ref(struct metadump_struct *md,
-			       struct extent_buffer *eb, int slot, int ext)
-{
-	struct btrfs_inode_extref *extref;
-	struct btrfs_inode_ref *ref;
-	char *garbage, *buf;
-	unsigned long ptr;
-	unsigned long name_ptr;
-	u32 item_size;
-	u32 cur_offset = 0;
-	int len;
-	int free_garbage = (md->sanitize_names == 1);
-
-	item_size = btrfs_item_size_nr(eb, slot);
-	ptr = btrfs_item_ptr_offset(eb, slot);
-	while (cur_offset < item_size) {
-		if (ext) {
-			extref = (struct btrfs_inode_extref *)(ptr +
-							       cur_offset);
-			name_ptr = (unsigned long)(&extref->name);
-			len = btrfs_inode_extref_name_len(eb, extref);
-			cur_offset += sizeof(*extref);
-		} else {
-			ref = (struct btrfs_inode_ref *)(ptr + cur_offset);
-			len = btrfs_inode_ref_name_len(eb, ref);
-			name_ptr = (unsigned long)(ref + 1);
-			cur_offset += sizeof(*ref);
-		}
-		cur_offset += len;
-
-		if (md->sanitize_names > 1) {
-			buf = malloc(len);
-			if (!buf) {
-				error("cannot sanitize name, not enough memory");
-				return;
-			}
-			read_extent_buffer(eb, buf, name_ptr, len);
-			garbage = find_collision(md, buf, len);
-		} else {
-			garbage = generate_garbage(len);
-		}
-
-		if (!garbage) {
-			error("cannot sanitize name, not enough memory");
-			return;
-		}
-		write_extent_buffer(eb, garbage, name_ptr, len);
-		if (free_garbage)
-			free(garbage);
-	}
-}
-
-static void sanitize_xattr(struct metadump_struct *md,
-			   struct extent_buffer *eb, int slot)
-{
-	struct btrfs_dir_item *dir_item;
-	unsigned long data_ptr;
-	u32 data_len;
-
-	dir_item = btrfs_item_ptr(eb, slot, struct btrfs_dir_item);
-	data_len = btrfs_dir_data_len(eb, dir_item);
-
-	data_ptr = (unsigned long)((char *)(dir_item + 1) +
-				   btrfs_dir_name_len(eb, dir_item));
-	memset_extent_buffer(eb, 0, data_ptr, data_len);
-}
-
-static void sanitize_name(struct metadump_struct *md, u8 *dst,
-			  struct extent_buffer *src, struct btrfs_key *key,
-			  int slot)
-{
-	struct extent_buffer *eb;
-
-	eb = alloc_dummy_eb(src->start, src->len);
-	if (!eb) {
-		error("cannot sanitize name, not enough memory");
-		return;
-	}
-
-	memcpy(eb->data, src->data, src->len);
-
-	switch (key->type) {
-	case BTRFS_DIR_ITEM_KEY:
-	case BTRFS_DIR_INDEX_KEY:
-		sanitize_dir_item(md, eb, slot);
-		break;
-	case BTRFS_INODE_REF_KEY:
-		sanitize_inode_ref(md, eb, slot, 0);
-		break;
-	case BTRFS_INODE_EXTREF_KEY:
-		sanitize_inode_ref(md, eb, slot, 1);
-		break;
-	case BTRFS_XATTR_ITEM_KEY:
-		sanitize_xattr(md, eb, slot);
-		break;
-	default:
-		break;
-	}
-
-	memcpy(dst, eb->data, eb->len);
-	free(eb);
-}
-
 /*
  * zero inline extents and csum items
  */
@@ -600,7 +284,8 @@ static void zero_items(struct metadump_struct *md, u8 *dst,
 		}
 
 		if (md->sanitize_names && has_name(&key)) {
-			sanitize_name(md, dst, src, &key, i);
+			sanitize_name(md->sanitize_names, &md->name_tree, dst,
+					src, &key, i);
 			continue;
 		}
 
@@ -746,7 +431,7 @@ static void metadump_destroy(struct metadump_struct *md, int num_threads)
 
 static int metadump_init(struct metadump_struct *md, struct btrfs_root *root,
 			 FILE *out, int num_threads, int compress_level,
-			 int sanitize_names)
+			 enum sanitize_mode sanitize_names)
 {
 	int i, ret = 0;
 
@@ -758,7 +443,7 @@ static int metadump_init(struct metadump_struct *md, struct btrfs_root *root,
 	md->pending_start = (u64)-1;
 	md->compress_level = compress_level;
 	md->sanitize_names = sanitize_names;
-	if (sanitize_names > 1)
+	if (sanitize_names == SANITIZE_COLLISIONS)
 		crc32c_optimization_init();
 
 	md->name_tree.rb_node = NULL;
@@ -814,7 +499,8 @@ static int write_buffers(struct metadump_struct *md, u64 *next)
 	}
 
 	if (err) {
-		error("one of the threads failed: %s", strerror(-err));
+		errno = -err;
+		error("one of the threads failed: %m");
 		goto out;
 	}
 
@@ -829,7 +515,7 @@ static int write_buffers(struct metadump_struct *md, u64 *next)
 
 	ret = fwrite(&md->cluster, BLOCK_SIZE, 1, md->out);
 	if (ret != 1) {
-		error("unable to write out cluster: %s", strerror(errno));
+		error("unable to write out cluster: %m");
 		return -errno;
 	}
 
@@ -845,8 +531,7 @@ static int write_buffers(struct metadump_struct *md, u64 *next)
 			ret = fwrite(async->buffer, async->bufsize, 1,
 				     md->out);
 		if (ret != 1) {
-			error("unable to write out cluster: %s",
-				strerror(errno));
+			error("unable to write out cluster: %m");
 			err = -errno;
 			ret = 0;
 		}
@@ -862,8 +547,7 @@ static int write_buffers(struct metadump_struct *md, u64 *next)
 		bytenr += size;
 		ret = write_zero(md->out, size);
 		if (ret != 1) {
-			error("unable to zero out buffer: %s",
-				strerror(errno));
+			error("unable to zero out buffer: %m");
 			err = -errno;
 		}
 	}
@@ -888,7 +572,7 @@ static int read_data_extent(struct metadump_struct *md,
 	num_copies = btrfs_num_copies(root->fs_info, logical, bytes_left);
 
 	/* Try our best to read data, just like read_tree_block() */
-	for (cur_mirror = 0; cur_mirror < num_copies; cur_mirror++) {
+	for (cur_mirror = 1; cur_mirror <= num_copies; cur_mirror++) {
 		while (bytes_left) {
 			read_len = bytes_left;
 			ret = read_extent_data(fs_info,
@@ -961,9 +645,8 @@ static int flush_pending(struct metadump_struct *md, int done)
 			if (ret < size) {
 				free(async->buffer);
 				free(async);
-				error("unable to read superblock at %llu: %s",
-						(unsigned long long)start,
-						strerror(errno));
+				error("unable to read superblock at %llu: %m",
+						(unsigned long long)start);
 				return -errno;
 			}
 			size = 0;
@@ -1008,10 +691,12 @@ static int flush_pending(struct metadump_struct *md, int done)
 	}
 	if (md->num_items >= ITEMS_PER_CLUSTER || done) {
 		ret = write_buffers(md, &start);
-		if (ret)
-			error("unable to write buffers: %s", strerror(-ret));
-		else
+		if (ret) {
+			errno = -ret;
+			error("unable to write buffers: %m");
+		} else {
 			meta_cluster_init(md, start);
+		}
 	}
 	pthread_mutex_unlock(&md->mutex);
 	return ret;
@@ -1035,43 +720,6 @@ static int add_extent(u64 start, u64 size, struct metadump_struct *md,
 	return 0;
 }
 
-#ifdef BTRFS_COMPAT_EXTENT_TREE_V0
-static int is_tree_block(struct btrfs_root *extent_root,
-			 struct btrfs_path *path, u64 bytenr)
-{
-	struct extent_buffer *leaf;
-	struct btrfs_key key;
-	u64 ref_objectid;
-	int ret;
-
-	leaf = path->nodes[0];
-	while (1) {
-		struct btrfs_extent_ref_v0 *ref_item;
-		path->slots[0]++;
-		if (path->slots[0] >= btrfs_header_nritems(leaf)) {
-			ret = btrfs_next_leaf(extent_root, path);
-			if (ret < 0)
-				return ret;
-			if (ret > 0)
-				break;
-			leaf = path->nodes[0];
-		}
-		btrfs_item_key_to_cpu(leaf, &key, path->slots[0]);
-		if (key.objectid != bytenr)
-			break;
-		if (key.type != BTRFS_EXTENT_REF_V0_KEY)
-			continue;
-		ref_item = btrfs_item_ptr(leaf, path->slots[0],
-					  struct btrfs_extent_ref_v0);
-		ref_objectid = btrfs_ref_objectid_v0(leaf, ref_item);
-		if (ref_objectid < BTRFS_FIRST_FREE_OBJECTID)
-			return 1;
-		break;
-	}
-	return 0;
-}
-#endif
-
 static int copy_tree_blocks(struct btrfs_root *root, struct extent_buffer *eb,
 			    struct metadump_struct *metadump, int root_tree)
 {
@@ -1289,30 +937,10 @@ static int copy_from_extent_tree(struct metadump_struct *metadump,
 				}
 			}
 		} else {
-#ifdef BTRFS_COMPAT_EXTENT_TREE_V0
-			ret = is_tree_block(extent_root, path, bytenr);
-			if (ret < 0) {
-				error("failed to check tree block %llu: %d",
-					(unsigned long long)bytenr, ret);
-				break;
-			}
-
-			if (ret) {
-				ret = add_extent(bytenr, num_bytes, metadump,
-						 0);
-				if (ret) {
-					error("unable to add block %llu: %d",
-						(unsigned long long)bytenr, ret);
-					break;
-				}
-			}
-			ret = 0;
-#else
 			error(
-	"either extent tree is corrupted or you haven't built with V0 support");
+	"either extent tree is corrupted or deprecated extent ref format");
 			ret = -EIO;
 			break;
-#endif
 		}
 		bytenr += num_bytes;
 	}
@@ -1323,7 +951,8 @@ static int copy_from_extent_tree(struct metadump_struct *metadump,
 }
 
 static int create_metadump(const char *input, FILE *out, int num_threads,
-			   int compress_level, int sanitize, int walk_trees)
+			   int compress_level, enum sanitize_mode sanitize,
+			   int walk_trees)
 {
 	struct btrfs_root *root;
 	struct btrfs_path path;
@@ -1655,8 +1284,7 @@ static void write_backup_supers(int fd, u8 *buf)
 
 	if (fstat(fd, &st)) {
 		error(
-	"cannot stat restore point, won't be able to write backup supers: %s",
-			strerror(errno));
+	"cannot stat restore point, won't be able to write backup supers: %m");
 		return;
 	}
 
@@ -1672,8 +1300,7 @@ static void write_backup_supers(int fd, u8 *buf)
 		if (ret < BTRFS_SUPER_INFO_SIZE) {
 			if (ret < 0)
 				error(
-				"problem writing out backup super block %d: %s",
-						i, strerror(errno));
+				"problem writing out backup super block %d: %m", i);
 			else
 				error("short write writing out backup super block");
 			break;
@@ -1720,7 +1347,7 @@ static void *restore_worker(void *data)
 		list_del_init(&async->list);
 
 		if (mdres->compress_method == COMPRESS_ZLIB) {
-			size = compress_size; 
+			size = compress_size;
 			pthread_mutex_unlock(&mdres->mutex);
 			ret = uncompress(buffer, (unsigned long *)&size,
 					 async->buffer, async->bufsize);
@@ -1781,8 +1408,7 @@ static void *restore_worker(void *data)
 
 error:
 				if (ret < 0) {
-					error("unable to write to device: %s",
-							strerror(errno));
+					error("unable to write to device: %m");
 					err = errno;
 				} else {
 					error("short write");
@@ -1913,9 +1539,12 @@ static int fill_mdres_info(struct mdrestore_struct *mdres,
 
 	super = (struct btrfs_super_block *)outbuf;
 	mdres->nodesize = btrfs_super_nodesize(super);
-	memcpy(mdres->fsid, super->fsid, BTRFS_FSID_SIZE);
-	memcpy(mdres->uuid, super->dev_item.uuid,
-		       BTRFS_UUID_SIZE);
+	if (btrfs_super_incompat_flags(super) &
+	    BTRFS_FEATURE_INCOMPAT_METADATA_UUID)
+		memcpy(mdres->fsid, super->metadata_uuid, BTRFS_FSID_SIZE);
+	else
+		memcpy(mdres->fsid, super->fsid, BTRFS_FSID_SIZE);
+	memcpy(mdres->uuid, super->dev_item.uuid, BTRFS_UUID_SIZE);
 	mdres->devid = le64_to_cpu(super->dev_item.devid);
 	free(buffer);
 	return 0;
@@ -1954,7 +1583,7 @@ static int add_cluster(struct meta_cluster *cluster,
 		}
 		ret = fread(async->buffer, async->bufsize, 1, mdres->in);
 		if (ret != 1) {
-			error("unable to read buffer: %s", strerror(errno));
+			error("unable to read buffer: %m");
 			free(async->buffer);
 			free(async);
 			return -EIO;
@@ -1984,7 +1613,7 @@ static int add_cluster(struct meta_cluster *cluster,
 		bytenr += size;
 		ret = fread(buffer, size, 1, mdres->in);
 		if (ret != 1) {
-			error("failed to read buffer: %s", strerror(errno));
+			error("failed to read buffer: %m");
 			return -EIO;
 		}
 	}
@@ -2042,7 +1671,7 @@ static int read_chunk_block(struct mdrestore_struct *mdres, u8 *buffer,
 
 	if (memcmp(mdres->fsid, eb->data + offsetof(struct btrfs_header, fsid),
 		   BTRFS_FSID_SIZE)) {
-		error("filesystem UUID of eb %llu does not match",
+		error("filesystem metadata UUID of eb %llu does not match",
 				(unsigned long long)bytenr);
 		ret = -EIO;
 		goto out;
@@ -2162,7 +1791,7 @@ static int search_for_chunk_blocks(struct mdrestore_struct *mdres,
 	bytenr = current_cluster;
 	while (1) {
 		if (fseek(mdres->in, current_cluster, SEEK_SET)) {
-			error("seek failed: %s", strerror(errno));
+			error("seek failed: %m");
 			ret = -EIO;
 			break;
 		}
@@ -2181,9 +1810,8 @@ static int search_for_chunk_blocks(struct mdrestore_struct *mdres,
 			ret = -EIO;
 			break;
 		} else if (ret < 0) {
-			error("unable to read image at %llu: %s",
-					(unsigned long long)cluster_bytenr,
-					strerror(errno));
+			error("unable to read image at %llu: %m",
+					(unsigned long long)cluster_bytenr);
 			break;
 		}
 		ret = 0;
@@ -2215,7 +1843,7 @@ static int search_for_chunk_blocks(struct mdrestore_struct *mdres,
 			if (mdres->compress_method == COMPRESS_ZLIB) {
 				ret = fread(tmp, bufsize, 1, mdres->in);
 				if (ret != 1) {
-					error("read error: %s", strerror(errno));
+					error("read error: %m");
 					ret = -EIO;
 					break;
 				}
@@ -2233,8 +1861,7 @@ static int search_for_chunk_blocks(struct mdrestore_struct *mdres,
 			} else {
 				ret = fread(buffer, bufsize, 1, mdres->in);
 				if (ret != 1) {
-					error("read error: %s",
-							strerror(errno));
+					error("read error: %m");
 					ret = -EIO;
 					break;
 				}
@@ -2287,7 +1914,7 @@ static int build_chunk_tree(struct mdrestore_struct *mdres,
 
 	ret = fread(cluster, BLOCK_SIZE, 1, mdres->in);
 	if (ret <= 0) {
-		error("unable to read cluster: %s", strerror(errno));
+		error("unable to read cluster: %m");
 		return -EIO;
 	}
 	ret = 0;
@@ -2309,7 +1936,7 @@ static int build_chunk_tree(struct mdrestore_struct *mdres,
 			break;
 		bytenr += le32_to_cpu(item->size);
 		if (fseek(mdres->in, le32_to_cpu(item->size), SEEK_CUR)) {
-			error("seek failed: %s", strerror(errno));
+			error("seek failed: %m");
 			return -EIO;
 		}
 	}
@@ -2328,7 +1955,7 @@ static int build_chunk_tree(struct mdrestore_struct *mdres,
 
 	ret = fread(buffer, le32_to_cpu(item->size), 1, mdres->in);
 	if (ret != 1) {
-		error("unable to read buffer: %s", strerror(errno));
+		error("unable to read buffer: %m");
 		free(buffer);
 		return -EIO;
 	}
@@ -2358,9 +1985,13 @@ static int build_chunk_tree(struct mdrestore_struct *mdres,
 	super = (struct btrfs_super_block *)buffer;
 	chunk_root_bytenr = btrfs_super_chunk_root(super);
 	mdres->nodesize = btrfs_super_nodesize(super);
-	memcpy(mdres->fsid, super->fsid, BTRFS_FSID_SIZE);
-	memcpy(mdres->uuid, super->dev_item.uuid,
-		       BTRFS_UUID_SIZE);
+	if (btrfs_super_incompat_flags(super) &
+	    BTRFS_FEATURE_INCOMPAT_METADATA_UUID)
+		memcpy(mdres->fsid, super->metadata_uuid, BTRFS_FSID_SIZE);
+	else
+		memcpy(mdres->fsid, super->fsid, BTRFS_FSID_SIZE);
+
+	memcpy(mdres->uuid, super->dev_item.uuid, BTRFS_UUID_SIZE);
 	mdres->devid = le64_to_cpu(super->dev_item.devid);
 	free(buffer);
 	pthread_mutex_unlock(&mdres->mutex);
@@ -2403,42 +2034,86 @@ static void remap_overlapping_chunks(struct mdrestore_struct *mdres)
 	}
 }
 
-static int fixup_devices(struct btrfs_fs_info *fs_info,
-			 struct mdrestore_struct *mdres, off_t dev_size)
+static int fixup_device_size(struct btrfs_trans_handle *trans,
+			     struct mdrestore_struct *mdres, int out_fd)
 {
-	struct btrfs_trans_handle *trans;
+	struct btrfs_fs_info *fs_info = trans->fs_info;
 	struct btrfs_dev_item *dev_item;
+	struct btrfs_dev_extent *dev_ext;
 	struct btrfs_path path;
 	struct extent_buffer *leaf;
 	struct btrfs_root *root = fs_info->chunk_root;
 	struct btrfs_key key;
+	struct stat buf;
 	u64 devid, cur_devid;
+	u64 dev_size; /* Get from last dev extents */
 	int ret;
 
-	trans = btrfs_start_transaction(fs_info->tree_root, 1);
-	if (IS_ERR(trans)) {
-		error("cannot starting transaction %ld", PTR_ERR(trans));
-		return PTR_ERR(trans);
-	}
-
 	dev_item = &fs_info->super_copy->dev_item;
 
+	btrfs_init_path(&path);
 	devid = btrfs_stack_device_id(dev_item);
 
+	key.objectid = devid;
+	key.type = BTRFS_DEV_EXTENT_KEY;
+	key.offset = (u64)-1;
+
+	ret = btrfs_search_slot(NULL, fs_info->dev_root, &key, &path, 0, 0);
+	if (ret < 0) {
+		errno = -ret;
+		error("failed to locate last dev extent of devid %llu: %m",
+			devid);
+		btrfs_release_path(&path);
+		return ret;
+	}
+	if (ret == 0) {
+		error("found invalid dev extent devid %llu offset -1", devid);
+		btrfs_release_path(&path);
+		return -EUCLEAN;
+	}
+	ret = btrfs_previous_item(fs_info->dev_root, &path, devid,
+				  BTRFS_DEV_EXTENT_KEY);
+	if (ret > 0)
+		ret = -ENOENT;
+	if (ret < 0) {
+		errno = -ret;
+		error("failed to locate last dev extent of devid %llu: %m",
+			devid);
+		btrfs_release_path(&path);
+		return ret;
+	}
+
+	btrfs_item_key_to_cpu(path.nodes[0], &key, path.slots[0]);
+	dev_ext = btrfs_item_ptr(path.nodes[0], path.slots[0],
+				 struct btrfs_dev_extent);
+	dev_size = key.offset + btrfs_dev_extent_length(path.nodes[0], dev_ext);
+	btrfs_release_path(&path);
+
 	btrfs_set_stack_device_total_bytes(dev_item, dev_size);
 	btrfs_set_stack_device_bytes_used(dev_item, mdres->alloced_chunks);
+	ret = fstat(out_fd, &buf);
+	if (ret < 0) {
+		error("failed to stat result image: %m");
+		return -errno;
+	}
+	if (S_ISREG(buf.st_mode)) {
+		/* Don't forget to enlarge the real file */
+		ret = ftruncate64(out_fd, dev_size);
+		if (ret < 0) {
+			error("failed to enlarge result image: %m");
+			return -errno;
+		}
+	}
 
 	key.objectid = BTRFS_DEV_ITEMS_OBJECTID;
 	key.type = BTRFS_DEV_ITEM_KEY;
 	key.offset = 0;
 
-	btrfs_init_path(&path);
-
 again:
 	ret = btrfs_search_slot(trans, root, &key, &path, -1, 1);
 	if (ret < 0) {
 		error("search failed: %d", ret);
-		exit(1);
+		return ret;
 	}
 
 	while (1) {
@@ -2485,12 +2160,192 @@ again:
 	}
 
 	btrfs_release_path(&path);
+	return 0;
+}
+
+static void fixup_block_groups(struct btrfs_fs_info *fs_info)
+{
+	struct btrfs_block_group_cache *bg;
+	struct btrfs_mapping_tree *map_tree = &fs_info->mapping_tree;
+	struct cache_extent *ce;
+	struct map_lookup *map;
+	u64 extra_flags;
+
+	for (ce = search_cache_extent(&map_tree->cache_tree, 0); ce;
+	     ce = next_cache_extent(ce)) {
+		map = container_of(ce, struct map_lookup, ce);
+
+		bg = btrfs_lookup_block_group(fs_info, ce->start);
+		if (!bg) {
+			warning(
+		"cannot find block group %llu, filesystem may not be mountable",
+				ce->start);
+			continue;
+		}
+		extra_flags = map->type & BTRFS_BLOCK_GROUP_PROFILE_MASK;
+
+		if (bg->flags == map->type)
+			continue;
+
+		/* Update the block group item and mark the bg dirty */
+		bg->flags = map->type;
+		btrfs_set_block_group_flags(&bg->item, bg->flags);
+		set_extent_bits(&fs_info->block_group_cache, ce->start,
+				ce->start + ce->size - 1, BLOCK_GROUP_DIRTY);
+
+		/*
+		 * Chunk and bg flags can be different, changing bg flags
+		 * without update avail_data/meta_alloc_bits will lead to
+		 * ENOSPC.
+		 * So here we set avail_*_alloc_bits to match chunk types.
+		 */
+		if (map->type & BTRFS_BLOCK_GROUP_DATA)
+			fs_info->avail_data_alloc_bits = extra_flags;
+		if (map->type & BTRFS_BLOCK_GROUP_METADATA)
+			fs_info->avail_metadata_alloc_bits = extra_flags;
+		if (map->type & BTRFS_BLOCK_GROUP_SYSTEM)
+			fs_info->avail_system_alloc_bits = extra_flags;
+	}
+}
+
+static int remove_all_dev_extents(struct btrfs_trans_handle *trans)
+{
+	struct btrfs_fs_info *fs_info = trans->fs_info;
+	struct btrfs_root *root = fs_info->dev_root;
+	struct btrfs_path path;
+	struct btrfs_key key;
+	struct extent_buffer *leaf;
+	int slot;
+	int ret;
+
+	key.objectid = 1;
+	key.type = BTRFS_DEV_EXTENT_KEY;
+	key.offset = 0;
+	btrfs_init_path(&path);
+
+	ret = btrfs_search_slot(trans, root, &key, &path, -1, 1);
+	if (ret < 0) {
+		errno = -ret;
+		error("failed to search dev tree: %m");
+		return ret;
+	}
+
+	while (1) {
+		slot = path.slots[0];
+		leaf = path.nodes[0];
+		if (slot >= btrfs_header_nritems(leaf)) {
+			ret = btrfs_next_leaf(root, &path);
+			if (ret < 0) {
+				errno = -ret;
+				error("failed to search dev tree: %m");
+				goto out;
+			}
+			if (ret > 0) {
+				ret = 0;
+				goto out;
+			}
+		}
+
+		btrfs_item_key_to_cpu(leaf, &key, slot);
+		if (key.type != BTRFS_DEV_EXTENT_KEY)
+			break;
+		ret = btrfs_del_item(trans, root, &path);
+		if (ret < 0) {
+			errno = -ret;
+			error("failed to delete dev extent %llu, %llu: %m",
+				key.objectid, key.offset);
+			goto out;
+		}
+	}
+out:
+	btrfs_release_path(&path);
+	return ret;
+}
+
+static int fixup_dev_extents(struct btrfs_trans_handle *trans)
+{
+	struct btrfs_fs_info *fs_info = trans->fs_info;
+	struct btrfs_mapping_tree *map_tree = &fs_info->mapping_tree;
+	struct btrfs_device *dev;
+	struct cache_extent *ce;
+	struct map_lookup *map;
+	u64 devid = btrfs_stack_device_id(&fs_info->super_copy->dev_item);
+	int i;
+	int ret;
+
+	ret = remove_all_dev_extents(trans);
+	if (ret < 0) {
+		errno = -ret;
+		error("failed to remove all existing dev extents: %m");
+	}
+
+	dev = btrfs_find_device(fs_info, devid, NULL, NULL);
+	if (!dev) {
+		error("faild to find devid %llu", devid);
+		return -ENODEV;
+	}
+
+	/* Rebuild all dev extents using chunk maps */
+	for (ce = search_cache_extent(&map_tree->cache_tree, 0); ce;
+	     ce = next_cache_extent(ce)) {
+		u64 stripe_len;
+
+		map = container_of(ce, struct map_lookup, ce);
+		stripe_len = calc_stripe_length(map->type, ce->size,
+						map->num_stripes);
+		for (i = 0; i < map->num_stripes; i++) {
+			ret = btrfs_insert_dev_extent(trans, dev, ce->start,
+					stripe_len, map->stripes[i].physical);
+			if (ret < 0) {
+				errno = -ret;
+				error(
+				"failed to insert dev extent %llu %llu: %m",
+					devid, map->stripes[i].physical);
+				goto out;
+			}
+		}
+	}
+out:
+	return ret;
+}
+
+static int fixup_chunks_and_devices(struct btrfs_fs_info *fs_info,
+			 struct mdrestore_struct *mdres, int out_fd)
+{
+	struct btrfs_trans_handle *trans;
+	int ret;
+
+	if (btrfs_super_log_root(fs_info->super_copy)) {
+		warning(
+		"log tree detected, its generation will not match superblock");
+	}
+	trans = btrfs_start_transaction(fs_info->tree_root, 1);
+	if (IS_ERR(trans)) {
+		error("cannot start transaction %ld", PTR_ERR(trans));
+		return PTR_ERR(trans);
+	}
+
+	fixup_block_groups(fs_info);
+	ret = fixup_dev_extents(trans);
+	if (ret < 0)
+		goto error;
+
+	ret = fixup_device_size(trans, mdres, out_fd);
+	if (ret < 0)
+		goto error;
+
 	ret = btrfs_commit_transaction(trans, fs_info->tree_root);
 	if (ret) {
 		error("unable to commit transaction: %d", ret);
 		return ret;
 	}
 	return 0;
+error:
+	errno = -ret;
+	error(
+"failed to fix chunks and devices mapping, the fs may not be mountable: %m");
+	btrfs_abort_transaction(trans, ret);
+	return ret;
 }
 
 static int restore_metadump(const char *input, FILE *out, int old_restore,
@@ -2510,8 +2365,7 @@ static int restore_metadump(const char *input, FILE *out, int old_restore,
 	} else {
 		in = fopen(input, "r");
 		if (!in) {
-			error("unable to open metadump image: %s",
-					strerror(errno));
+			error("unable to open metadump image: %m");
 			return 1;
 		}
 	}
@@ -2552,7 +2406,7 @@ static int restore_metadump(const char *input, FILE *out, int old_restore,
 	}
 
 	if (in != stdin && fseek(in, 0, SEEK_SET)) {
-		error("seek failed: %s", strerror(errno));
+		error("seek failed: %m");
 		goto out;
 	}
 
@@ -2592,13 +2446,13 @@ static int restore_metadump(const char *input, FILE *out, int old_restore,
 		info = root->fs_info;
 
 		if (stat(target, &st)) {
-			error("stat %s failed: %s", target, strerror(errno));
+			error("stat %s failed: %m", target);
 			close_ctree(info->chunk_root);
 			free(cluster);
 			return 1;
 		}
 
-		ret = fixup_devices(info, &mdrestore, st.st_size);
+		ret = fixup_chunks_and_devices(info, &mdrestore, fileno(out));
 		close_ctree(info->chunk_root);
 		if (ret)
 			goto out;
@@ -2637,7 +2491,7 @@ static int update_disk_super_on_device(struct btrfs_fs_info *info,
 	key.offset = cur_devid;
 
 	btrfs_init_path(&path);
-	ret = btrfs_search_slot(NULL, info->chunk_root, &key, &path, 0, 0); 
+	ret = btrfs_search_slot(NULL, info->chunk_root, &key, &path, 0, 0);
 	if (ret) {
 		error("search key failed: %d", ret);
 		ret = -EIO;
@@ -2673,7 +2527,7 @@ static int update_disk_super_on_device(struct btrfs_fs_info *info,
 	/* update other devices' super block */
 	fp = open(other_dev, O_CREAT | O_RDWR, 0600);
 	if (fp < 0) {
-		error("could not open %s: %s", other_dev, strerror(errno));
+		error("could not open %s: %m", other_dev);
 		ret = -EIO;
 		goto out;
 	}
@@ -2696,10 +2550,12 @@ static int update_disk_super_on_device(struct btrfs_fs_info *info,
 
 	ret = pwrite64(fp, buf, BTRFS_SUPER_INFO_SIZE, BTRFS_SUPER_INFO_OFFSET);
 	if (ret != BTRFS_SUPER_INFO_SIZE) {
-		if (ret < 0)
-			error("cannot write superblock: %s", strerror(ret));
-		else
+		if (ret < 0) {
+			errno = ret;
+			error("cannot write superblock: %m");
+		} else {
 			error("cannot write superblock");
+		}
 		ret = -EIO;
 		goto out;
 	}
@@ -2739,7 +2595,7 @@ int main(int argc, char *argv[])
 	int walk_trees = 0;
 	int multi_devices = 0;
 	int ret;
-	int sanitize = 0;
+	enum sanitize_mode sanitize = SANITIZE_NONE;
 	int dev_cnt = 0;
 	int usage_error = 0;
 	FILE *out;
@@ -2777,7 +2633,10 @@ int main(int argc, char *argv[])
 			old_restore = 1;
 			break;
 		case 's':
-			sanitize++;
+			if (sanitize == SANITIZE_NONE)
+				sanitize = SANITIZE_NAMES;
+			else if (sanitize == SANITIZE_NAMES)
+				sanitize = SANITIZE_COLLISIONS;
 			break;
 		case 'w':
 			walk_trees = 1;
@@ -2805,9 +2664,9 @@ int main(int argc, char *argv[])
 			usage_error++;
 		}
 	} else {
-		if (walk_trees || sanitize || compress_level) {
+		if (walk_trees || sanitize != SANITIZE_NONE || compress_level) {
 			error(
-			"useing -w, -s, -c options for restore makes no sense");
+			"using -w, -s, -c options for restore makes no sense");
 			usage_error++;
 		}
 		if (multi_devices && dev_cnt < 2) {
@@ -2851,8 +2710,8 @@ int main(int argc, char *argv[])
 	if (create) {
 		ret = check_mounted(source);
 		if (ret < 0) {
-			warning("unable to check mount status of: %s",
-					strerror(-ret));
+			errno = -ret;
+			warning("unable to check mount status of: %m");
 		} else if (ret) {
 			warning("%s already mounted, results may be inaccurate",
 					source);
@@ -2865,8 +2724,7 @@ int main(int argc, char *argv[])
 				       0, target, multi_devices);
 	}
 	if (ret) {
-		error("%s failed: %s", (create) ? "create" : "restore",
-		       strerror(errno));
+		error("%s failed: %m", (create) ? "create" : "restore");
 		goto out;
 	}
 
@@ -2924,8 +2782,8 @@ out:
 
 			unlink_ret = unlink(target);
 			if (unlink_ret)
-				error("unlink output file %s failed: %s",
-						target, strerror(errno));
+				error("unlink output file %s failed: %m",
+						target);
 		}
 	}
 
